from NEMO.actions import has_perm
from NEMO.typing import QuerySetType
from NEMO.utilities import new_model_copy
from django import forms
from django.contrib import admin, messages
from django.urls import reverse
from django.utils.safestring import mark_safe

from NEMO_online_training.models import OnlineTraining, OnlineTrainingAction, OnlineUserTraining, ProspectiveUser
from NEMO_online_training.training_actions import action_handlers


@admin.action(description="Duplicate selected training")
def duplicate_online_training(model_admin, request, queryset: QuerySetType[OnlineTraining]):
    if not has_perm(request, queryset, "add") or not has_perm(request, queryset, "change"):
        model_admin.message_user(request, "You do not have permission to run this action.", level=messages.ERROR)
    for online_training in queryset:
        original_name = online_training.name
        new_name = "Copy of " + online_training.name
        try:
            if OnlineTraining.objects.filter(name=new_name).exists():
                messages.error(
                    request,
                    mark_safe(
                        f'There is already a copy of {original_name} as <a href="{reverse("admin:NEMO_online_training_onlinetraining_change", args=[online_training.id])}">{new_name}</a>. Change the copy\'s name and try again'
                    ),
                )
                continue
            else:
                old_actions = online_training.onlinetrainingaction_set.all()
                new_online_training = new_model_copy(online_training)
                new_online_training.name = new_name
                new_online_training.save()
                for action in old_actions:
                    new_online_training.onlinetrainingaction_set.add(action)
                messages.success(
                    request,
                    mark_safe(
                        f'A duplicate of {original_name} has been made as <a href="{reverse("admin:NEMO_online_training_onlinetraining_change", args=[new_online_training.id])}">{new_online_training.name}</a>'
                    ),
                )
        except Exception as error:
            messages.error(
                request, f"{original_name} could not be duplicated because of the following error: {str(error)}"
            )


@admin.register(ProspectiveUser)
class ProspectiveUserAdmin(admin.ModelAdmin):
    list_display = [
        "first_name",
        "last_name",
        "email",
        "get_all_blocking_trainings_completed",
        "get_all_trainings_completed",
        "last_accessed",
        "creation_time",
        "id",
    ]
    date_hierarchy = "creation_time"
    autocomplete_fields = ["nemo_user"]
    readonly_fields = ["creation_time", "last_updated", "last_accessed"]

    @admin.display(boolean=True, description="All Trainings Completed")
    def get_all_trainings_completed(self, obj: ProspectiveUser) -> bool:
        return obj.all_trainings_completed()

    @admin.display(boolean=True, description="All Blocking Trainings Completed")
    def get_all_blocking_trainings_completed(self, obj: ProspectiveUser) -> bool:
        return obj.all_blocking_trainings_completed()


class OnlineTrainingActionInlineForm(forms.ModelForm):
    action_type = forms.ChoiceField(choices=[])  # Start empty

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields["action_type"].choices = [(action.name, action.description) for action in action_handlers.values()]

    class Meta:
        model = OnlineTrainingAction
        fields = "__all__"


class OnlineTrainingActionInline(admin.TabularInline):
    model = OnlineTrainingAction
    extra = 0
    verbose_name = "After training is completed"
    verbose_name_plural = "After training is completed"
    form = OnlineTrainingActionInlineForm


@admin.register(OnlineTraining)
class OnlineTrainingAdmin(admin.ModelAdmin):
    inlines = [OnlineTrainingActionInline]
    list_display = ["name", "enabled", "is_blocking", "completion_time_limit", "creation_time", "id"]
    date_hierarchy = "creation_time"
    list_filter = ["enabled", "is_blocking"]
    actions = [duplicate_online_training]


@admin.register(OnlineUserTraining)
class OnlineUserTrainingAdmin(admin.ModelAdmin):
    list_display = [
        "prospective_user",
        "online_training",
        "get_training_completed",
        "get_training_expired",
        "due_date",
        "completion_time",
        "start",
        "end",
        "creation_time",
        "last_updated",
        "id",
    ]
    list_filter = ["online_training"]
    date_hierarchy = "creation_time"
    readonly_fields = ["creation_time", "last_updated"]

    @admin.display(boolean=True, description="Completed")
    def get_training_completed(self, obj: OnlineUserTraining) -> bool:
        return obj.completed()

    @admin.display(boolean=True, description="Expired")
    def get_training_expired(self, obj: OnlineUserTraining) -> bool:
        return obj.has_training_expired()



# Generated by Django 4.2.29 on 2026-03-11 14:36

import NEMO_online_training.fields
from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="OnlineTraining",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("name", models.CharField(max_length=255, unique=True)),
                (
                    "enabled",
                    models.BooleanField(
                        default=True, help_text="If unchecked, this training will not be available to users"
                    ),
                ),
                (
                    "completion_time_limit",
                    models.PositiveIntegerField(
                        default=120,
                        help_text="The maximum time the user can take to complete the training on the page (in minutes)",
                    ),
                ),
                (
                    "is_blocking",
                    models.BooleanField(
                        default=False,
                        help_text="If checked, prevents the linked NEMO user from performing other actions until this training is completed",
                    ),
                ),
                (
                    "allow_self_enrollment",
                    models.BooleanField(
                        default=False, help_text="Allow users to take this training without prior assignment."
                    ),
                ),
                (
                    "html_content",
                    models.TextField(
                        blank=True,
                        help_text="The HTML content of the training. The following context variables are available:<ul style='padding-left: 35px;'><li style='list-style: initial'><b>training_user</b>: the user</li><li style='list-style: initial'><b>training</b>: this training</li><li style='list-style: initial'><b>record</b>: the completion record</li></ul>Upon completion, call the JS function: <code>training_completed(dict_data)</code>",
                        null=True,
                    ),
                ),
                ("creation_time", models.DateTimeField(auto_now_add=True)),
            ],
            options={
                "ordering": ["name"],
            },
        ),
        migrations.CreateModel(
            name="ProspectiveUser",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("creation_time", models.DateTimeField(auto_now_add=True)),
                ("last_updated", models.DateTimeField(auto_now=True)),
                ("last_accessed", models.DateTimeField(blank=True, null=True)),
                (
                    "_first_name",
                    models.CharField(
                        blank=True, db_column="first_name", max_length=100, null=True, verbose_name="First name"
                    ),
                ),
                (
                    "_last_name",
                    models.CharField(
                        blank=True, db_column="last_name", max_length=100, null=True, verbose_name="Last name"
                    ),
                ),
                (
                    "_email",
                    models.EmailField(
                        blank=True, db_column="email", max_length=254, null=True, verbose_name="Email address"
                    ),
                ),
                (
                    "nemo_user",
                    models.OneToOneField(
                        blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to=settings.AUTH_USER_MODEL
                    ),
                ),
            ],
            options={
                "ordering": ["_first_name", "_last_name", "nemo_user__first_name", "nemo_user__last_name"],
            },
        ),
        migrations.CreateModel(
            name="OnlineUserTraining",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                (
                    "due_date",
                    models.DateTimeField(blank=True, help_text="The due date/time for the training", null=True),
                ),
                (
                    "start",
                    models.DateTimeField(blank=True, help_text="The date/time the training was started", null=True),
                ),
                (
                    "end",
                    models.DateTimeField(blank=True, help_text="The date/time the training was completed", null=True),
                ),
                (
                    "completion_data",
                    models.JSONField(
                        blank=True,
                        help_text="Completion data to be stored for the training, like answers to questions",
                        null=True,
                    ),
                ),
                ("creation_time", models.DateTimeField(auto_now_add=True)),
                ("last_updated", models.DateTimeField(auto_now=True)),
                (
                    "online_training",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, to="NEMO_online_training.onlinetraining"
                    ),
                ),
                (
                    "prospective_user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, to="NEMO_online_training.prospectiveuser"
                    ),
                ),
            ],
            options={
                "ordering": ["-end", "-due_date"],
            },
        ),
        migrations.CreateModel(
            name="OnlineTrainingAction",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("action_type", models.CharField(max_length=100)),
                (
                    "configuration",
                    models.JSONField(
                        blank=True,
                        default=dict,
                        help_text="Enter configuration as JSON. See <a href='https://github.com/usnistgov/NEMO-online-training?tab=readme-ov-file#usage' target='_blank'>this link</a> for more information.",
                    ),
                ),
                (
                    "user_filter",
                    NEMO_online_training.fields.UserTypeFilterField(
                        default="all_nemo,prospective",
                        help_text="Select which users this action applies to. You can select 'All NEMO users', specific user types, and/or new users (without NEMO accounts).",
                    ),
                ),
                (
                    "online_training",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, to="NEMO_online_training.onlinetraining"
                    ),
                ),
            ],
            options={
                "ordering": ["action_type"],
            },
        ),
    ]



from django.apps import AppConfig
from django.conf import settings


class OnlineTrainingConfig(AppConfig):
    name = "NEMO_online_training"
    verbose_name = "online training"
    plugin_id = 2200
    default_auto_field = "django.db.models.AutoField"

    @classmethod
    def get_plugin_id(cls):
        # Used to make EmailCategory and other IntegerChoices ranges unique
        return getattr(settings, f"{cls.name.upper()}_PLUGIN_ID", cls.plugin_id)

    def ready(self):
        from django.utils.translation import gettext_lazy as _
        from NEMO.plugins.utils import (
            add_dynamic_email_categories,
            add_dynamic_notification_types,
            add_extra_policy_class,
            check_extra_dependencies,
        )
        from NEMO_online_training.utilities import (
            ONLINE_TRAINING_EMAIL_CATEGORY,
            ONLINE_TRAINING_NOTIFICATION_TYPE,
        )
        from NEMO_online_training.policy import NEMOOnlineTrainingPolicy

        check_extra_dependencies(self.name, ["NEMO", "NEMO-CE"])

        add_dynamic_email_categories([(ONLINE_TRAINING_EMAIL_CATEGORY, _("Online training"))])
        add_dynamic_notification_types(
            [(ONLINE_TRAINING_NOTIFICATION_TYPE, "New online trainings - notifies users with pending trainings")]
        )
        add_extra_policy_class(f"{NEMOOnlineTrainingPolicy.__module__}.{NEMOOnlineTrainingPolicy.__qualname__}")



from NEMO.serializers import ModelSerializer
from NEMO.views.api import ModelViewSet, boolean_filters, datetime_filters, key_filters, number_filters, string_filters
from rest_flex_fields.serializers import FlexFieldsSerializerMixin

from NEMO_online_training.models import OnlineTraining, OnlineTrainingAction, OnlineUserTraining, ProspectiveUser


class OnlineTrainingSerializer(ModelSerializer):
    class Meta:
        model = OnlineTraining
        fields = "__all__"


class ProspectiveUserSerializer(FlexFieldsSerializerMixin, ModelSerializer):
    class Meta:
        model = ProspectiveUser
        fields = "__all__"
        expandable_fields = {
            "nemo_user": "NEMO.serializers.UserSerializer",
        }


class OnlineUserTrainingSerializer(FlexFieldsSerializerMixin, ModelSerializer):
    class Meta:
        model = OnlineUserTraining
        fields = "__all__"
        expandable_fields = {
            "online_training": "NEMO_online_training.api.OnlineTrainingSerializer",
            "prospective_user": "NEMO_online_training.api.ProspectiveUserSerializer",
        }


class OnlineTrainingActionSerializer(FlexFieldsSerializerMixin, ModelSerializer):
    class Meta:
        model = OnlineTrainingAction
        fields = "__all__"
        expandable_fields = {
            "online_training": "NEMO_online_training.api.OnlineTrainingSerializer",
        }


class OnlineTrainingViewSet(ModelViewSet):
    filename = "online_trainings"
    queryset = OnlineTraining.objects.all()
    serializer_class = OnlineTrainingSerializer
    filterset_fields = {
        "id": key_filters,
        "name": string_filters,
        "enabled": boolean_filters,
        "completion_time_limit": number_filters,
        "is_blocking": boolean_filters,
        "allow_self_enrollment": boolean_filters,
        "html_content": string_filters,
        "creation_time": datetime_filters,
    }


class ProspectiveUserViewSet(ModelViewSet):
    filename = "prospective_users"
    queryset = ProspectiveUser.objects.all()
    serializer_class = ProspectiveUserSerializer
    filterset_fields = {
        "id": key_filters,
        "creation_time": datetime_filters,
        "last_updated": datetime_filters,
        "last_accessed": datetime_filters,
        "_first_name": string_filters,
        "_last_name": string_filters,
        "_email": string_filters,
        "nemo_user": key_filters,
    }


class OnlineUserTrainingViewSet(ModelViewSet):
    filename = "online_user_trainings"
    queryset = OnlineUserTraining.objects.all()
    serializer_class = OnlineUserTrainingSerializer
    filterset_fields = {
        "id": key_filters,
        "online_training": key_filters,
        "prospective_user": key_filters,
        "due_date": datetime_filters,
        "start": datetime_filters,
        "end": datetime_filters,
        "completion_data": [],
        "creation_time": datetime_filters,
        "last_updated": datetime_filters,
    }


class OnlineTrainingActionViewSet(ModelViewSet):
    filename = "online_training_actions"
    queryset = OnlineTrainingAction.objects.all()
    serializer_class = OnlineTrainingActionSerializer
    filterset_fields = {
        "id": key_filters,
        "online_training": key_filters,
        "action_type": string_filters,
        "configuration": [],
        "user_filter": string_filters,
    }



from NEMO.urls import router, sort_urls
from django.urls import include, path

from NEMO_online_training import api
from NEMO_online_training.views import online_training

# Rest API URLs
router.register(r"online_training/online_training_actions", api.OnlineTrainingActionViewSet)
router.register(r"online_training/online_trainings", api.OnlineTrainingViewSet)
router.register(r"online_training/online_user_trainings", api.OnlineUserTrainingViewSet)
router.register(r"online_training/prospective_users", api.ProspectiveUserViewSet)
router.registry.sort(key=sort_urls)

urlpatterns = [
    path(
        "online_training/",
        include(
            [
                path("user_trainings/", online_training.user_online_trainings, name="online_user_trainings"),
                path(
                    "user_trainings/user/<int:prospective_user_id>/",
                    online_training.user_online_trainings,
                    name="online_user_trainings",
                ),
                path(
                    "user_trainings/search_users/",
                    online_training.search_prospective_users,
                    name="online_training_search_users",
                ),
                path(
                    "user_trainings/search_results/",
                    online_training.prospective_users_search_results,
                    name="online_training_search_users_results",
                ),
                path(
                    "users/create_from_nemo_user/<int:nemo_user_id>/",
                    online_training.create_prospective_user_from_nemo_user,
                    name="online_training_create_prospective_user_from_nemo_user",
                ),
                path(
                    "users/create_nemo_user/<int:prospective_user_id>/",
                    online_training.create_nemo_user_from_prospective_user,
                    name="online_training_create_nemo_user_from_prospective_user",
                ),
                path(
                    "users/create_user/",
                    online_training.create_prospective_user,
                    name="online_training_create_prospective_user",
                ),
                path(
                    "user_trainings/<int:user_training_id>/",
                    online_training.training,
                    name="online_training_user_training",
                ),
                path(
                    "trainings/<int:online_training_id>/",
                    online_training.training_without_assignment,
                    name="online_training_training",
                ),
                path(
                    "user_trainings/<int:prospective_user_id>/<int:online_training_id>/add/",
                    online_training.add_training_to_user,
                    name="online_training_add_training_to_user",
                ),
                path(
                    "public/user_training/<str:signed_user_training_id>/",
                    online_training.public_user_training,
                    name="public_online_training_user_training",
                ),
                path(
                    "public/complete_user_training/",
                    online_training.public_complete_user_training,
                    name="public_online_training_complete_user_training",
                ),
                path(
                    "public/generate_user_training_email/<int:user_training_id>/",
                    online_training.public_generate_user_training_email,
                    name="public_online_training_generate_user_training_email",
                ),
            ]
        ),
    )
]



from typing import List, Tuple

from NEMO.fields import CommaSeparatedListConverter, CommaSeparatedTextMultipleChoiceField, DynamicChoicesTextField
from NEMO.models import UserType
from NEMO.utilities import safe_lazy_queryset_evaluation
from django.contrib.admin.widgets import FilteredSelectMultiple
from django.db import models
from django.utils.translation import gettext_lazy as _


class UserTypeFilterField(DynamicChoicesTextField):
    """
    A field for filtering users by type, similar to NEMO's MultiRoleGroupPermissionChoiceField.

    Allows selection of:
    - "all_nemo" - All NEMO users regardless of type
    - "prospective" - Prospective users without NEMO accounts
    - User type IDs - Specific NEMO user types

    Storage format: Comma-separated string (e.g., "all_nemo,prospective" or "1,3,prospective")
    """

    # Special filter values
    ALL_NEMO_USERS = "all_nemo"
    PROSPECTIVE_USERS = "prospective"

    # Display labels
    LABEL_ALL_NEMO = _("All NEMO users")
    LABEL_PROSPECTIVE = _("All New users")

    def user_type_choices(self) -> List[Tuple[str, str]]:
        """
        Generate the list of choices for the field.

        Returns:
            List of (value, label) tuples for the dropdown/checkboxes
        """
        choices = [
            (self.ALL_NEMO_USERS, str(self.LABEL_ALL_NEMO)),
            (self.PROSPECTIVE_USERS, str(self.LABEL_PROSPECTIVE)),
        ]

        user_types, error = safe_lazy_queryset_evaluation(UserType.objects.all().order_by("name"))
        for user_type in user_types:
            choices.append((str(user_type.id), user_type.name))
        return choices

    def formfield(self, **kwargs):
        """
        Override to provide custom widget (FilteredSelectMultiple) for admin interface.
        """
        choices = kwargs.pop("choices", self.user_type_choices())
        is_stacked = kwargs.pop("is_stacked", False)
        kwargs["widget"] = FilteredSelectMultiple(_("User Types"), is_stacked=is_stacked)
        return super(models.TextField, self).formfield(
            choices=choices, form_class=CommaSeparatedTextMultipleChoiceField, **kwargs
        )

    def to_python(self, value) -> List:
        """Convert stored CSV string to Python list."""
        return CommaSeparatedListConverter.to_list(value)

    def from_db_value(self, value, *args, **kwargs) -> List:
        """Convert database value to Python list."""
        return self.to_python(value)

    def get_prep_value(self, value) -> str:
        """Convert Python list to CSV string for database storage."""
        return CommaSeparatedListConverter.to_string(value)

    def value_from_object(self, obj):
        """Get the value from the object and prepare it for forms."""
        value = super().value_from_object(obj)
        return self.get_prep_value(value)

    def applies_to_user(self, filter_values: List[str], prospective_user) -> bool:
        """
        Check if this filter applies to the given prospective user.

        Args:
            filter_values: List of filter values (e.g., ["all_nemo", "prospective", "1", "3"])
            prospective_user: ProspectiveUser instance to check

        Returns:
            True if the filter applies to this user, False otherwise
        """
        if not filter_values:
            # Empty filter = applies to nobody
            return False

        # Check if the user has NEMO account
        if prospective_user.nemo_user:
            # User has NEMO account
            # Check if "all_nemo" is in the filter
            if self.ALL_NEMO_USERS in filter_values:
                return True

            # Check if the user's type ID is in the filter
            user_type_id = str(prospective_user.nemo_user.type_id)
            if user_type_id in filter_values:
                return True

            return False
        else:
            # User doesn't have NEMO account
            # Check if "prospective" is in the filter
            return self.PROSPECTIVE_USERS in filter_values

    def user_types_display(self, filter_values: List[str]) -> str:
        """
        Convert the list of selected filter values to a readable string for display.

        Args:
            filter_values: List of filter values

        Returns:
            Comma-separated string of labels
        """
        if not filter_values:
            return _("None")

        choices_dict = dict(self.user_type_choices())
        labels = []

        for value in filter_values:
            label = choices_dict.get(value, value)
            labels.append(label)

        return ", ".join(labels)

    def get_choices(self, *args, **kwargs):
        """Override to ensure blank is not included."""
        kwargs["include_blank"] = False
        return super().get_choices(*args, **kwargs)



from django import forms
from django.core.exceptions import ValidationError
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from NEMO_online_training.models import OnlineUserTraining, ProspectiveUser


# Using non-underscore fields here since django Forms don't like variables starting with _
class ProspectiveUserForm(forms.ModelForm):
    first_name = forms.CharField(label="First name", required=True)
    last_name = forms.CharField(label="Last name", required=True)
    email = forms.EmailField(label="Email", required=True)

    class Meta:
        model = ProspectiveUser
        exclude = ["_first_name", "_last_name", "_email"]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.instance and self.instance.pk:
            self.fields["first_name"].initial = self.instance._first_name
            self.fields["last_name"].initial = self.instance._last_name
            self.fields["email"].initial = self.instance._email

    def clean(self):
        cleaned_data = super().clean()
        self.instance._first_name = cleaned_data.get("first_name", "").strip()
        self.instance._last_name = cleaned_data.get("last_name", "").strip()
        self.instance._email = cleaned_data.get("email", "").strip()
        return cleaned_data


class OnlineUserTrainingForm(forms.ModelForm):
    class Meta:
        model = OnlineUserTraining
        fields = ["due_date"]

    def clean_due_date(self):
        due_date = self.cleaned_data.get("due_date")
        if due_date and due_date < timezone.now():
            raise ValidationError(_("Due date must be in the future."))
        return due_date



from datetime import timedelta

from NEMO.models import EmailLog, User, UserType
from NEMO.tests.test_utilities import NEMOTestCaseMixin
from django.apps import apps
from django.core.exceptions import ValidationError
from django.test import TestCase
from django.utils import timezone

from NEMO_online_training.models import OnlineTraining, OnlineTrainingAction, OnlineUserTraining, ProspectiveUser
from NEMO_online_training.training_actions import action_handlers
from NEMO_online_training.utilities import ONLINE_TRAINING_ACTION_EXTEND_ACCESS, ONLINE_TRAINING_ACTION_SEND_EMAIL


class OnlineTrainingTest(NEMOTestCaseMixin, TestCase):

    def setUp(self):
        # Create user types
        self.staff_type = UserType.objects.create(name="Staff", display_order=1)
        self.student_type = UserType.objects.create(name="Student", display_order=2)
        self.technician_type = UserType.objects.create(name="Technician", display_order=3)

        # Create NEMO users of different types
        self.staff_user = User.objects.create(
            username="staff",
            first_name="Staff",
            last_name="User",
            email="staff@nemo.com",
            type=self.staff_type,
        )
        self.student_user = User.objects.create(
            username="student",
            first_name="Student",
            last_name="User",
            email="student@nemo.com",
            type=self.student_type,
        )

        self.prospective_staff = ProspectiveUser.create_from_nemo_user(self.staff_user)
        self.prospective_student = ProspectiveUser.create_from_nemo_user(self.student_user)
        self.prospective_only = ProspectiveUser.objects.create(
            first_name="Prospective", last_name="User", email="prospective@nemo.com"
        )

        self.training = OnlineTraining.objects.create(name="Test Training", completion_time_limit=120, enabled=True)

    def test_plugin_is_installed(self):
        assert apps.is_installed("NEMO_online_training")

    def test_extend_access_all_nemo_users(self):
        """Test that extend access action applies to all NEMO users when 'all_nemo' is selected."""
        # Create action for all NEMO users
        action = OnlineTrainingAction.objects.create(
            online_training=self.training,
            action_type=ONLINE_TRAINING_ACTION_EXTEND_ACCESS,
            configuration={"extend_by_days": 365},
            user_filter="all_nemo",  # All NEMO users
        )
        action.full_clean()

        # Set initial access expiration dates
        initial_date = timezone.now()
        self.staff_user.access_expiration = initial_date
        self.student_user.access_expiration = initial_date
        self.staff_user.save()
        self.student_user.save()

        # Test that action applies to all NEMO users
        self.assertTrue(action.applies_to_user(self.prospective_staff))
        self.assertTrue(action.applies_to_user(self.prospective_student))
        self.assertFalse(action.applies_to_user(self.prospective_only))

        # Perform action for staff user
        user_training_staff = OnlineUserTraining.objects.create(
            online_training=self.training, prospective_user=self.prospective_staff
        )
        action_handlers[action.action_type].perform(action, user_training_staff)

        # Verify access was extended
        self.staff_user.refresh_from_db()
        expected_date = initial_date + timedelta(days=365)
        self.assertEqual(self.staff_user.access_expiration, expected_date.date())

        # Perform action for student user
        user_training_student = OnlineUserTraining.objects.create(
            online_training=self.training, prospective_user=self.prospective_student
        )
        action_handlers[action.action_type].perform(action, user_training_student)

        # Verify access was extended
        self.student_user.refresh_from_db()
        self.assertEqual(self.student_user.access_expiration, expected_date.date())

    def test_extend_access_specific_user_types(self):
        """Test that extend access action applies only to specific user types."""
        # Create action for staff and technician only
        action = OnlineTrainingAction.objects.create(
            online_training=self.training,
            action_type=ONLINE_TRAINING_ACTION_EXTEND_ACCESS,
            configuration={"extend_by_days": 180},
            user_filter=f"{self.staff_type.id},{self.technician_type.id}",  # Staff and Technician only
        )
        action.full_clean()

        # Set initial access expiration dates
        initial_date = timezone.now()
        self.staff_user.access_expiration = initial_date
        self.student_user.access_expiration = initial_date
        self.staff_user.save()
        self.student_user.save()

        # Test that action applies to staff and technician, but not student
        self.assertTrue(action.applies_to_user(self.prospective_staff))
        self.assertFalse(action.applies_to_user(self.prospective_student))
        self.assertFalse(action.applies_to_user(self.prospective_only))

        # Perform action for staff user
        user_training_staff = OnlineUserTraining.objects.create(
            online_training=self.training, prospective_user=self.prospective_staff
        )
        action_handlers[action.action_type].perform(action, user_training_staff)

        # Verify staff access was extended
        self.staff_user.refresh_from_db()
        expected_date = initial_date + timedelta(days=180)
        self.assertEqual(self.staff_user.access_expiration, expected_date.date())

        # Perform action for student user
        user_training_student = OnlineUserTraining.objects.create(
            online_training=self.training, prospective_user=self.prospective_student
        )
        action_handlers[action.action_type].perform(action, user_training_student)

        # Student's access should not have changed
        self.student_user.refresh_from_db()
        self.assertEqual(self.student_user.access_expiration, initial_date.date())

    def test_extend_access_prospective_users_only(self):
        """Test that extend access action only for prospective users fails."""
        action = OnlineTrainingAction.objects.create(
            online_training=self.training,
            action_type=ONLINE_TRAINING_ACTION_EXTEND_ACCESS,
            configuration={"extend_by_days": 90},
            user_filter="prospective",  # Prospective users only
        )
        # Should raise a validation error because prospective users are not allowed to have access expiration
        self.assertRaises(ValidationError, action.full_clean)

    def test_extend_access_combined_filter(self):
        """Test that extend access action applies to specific types AND prospective users."""
        # Create action for staff type + prospective users
        action = OnlineTrainingAction.objects.create(
            online_training=self.training,
            action_type=ONLINE_TRAINING_ACTION_EXTEND_ACCESS,
            configuration={"extend_by_days": 30},
            user_filter=f"{self.staff_type.id},prospective",  # Staff + prospective
        )
        # Should raise a validation error because prospective users are not allowed to have access expiration
        self.assertRaises(ValidationError, action.full_clean)

    def test_extend_access_empty_filter(self):
        """Test that action with an empty filter doesn't pass validation."""
        action = OnlineTrainingAction.objects.create(
            online_training=self.training,
            action_type=ONLINE_TRAINING_ACTION_EXTEND_ACCESS,
            configuration={"extend_by_days": 365},
            user_filter="",  # Empty filter
        )
        # Should raise a validation error because an empty filter is not allowed
        self.assertRaises(ValidationError, action.full_clean)

    def test_email_prospective_users_only(self):
        """Test that email access action applies only to prospective users."""
        action = OnlineTrainingAction.objects.create(
            online_training=self.training,
            action_type=ONLINE_TRAINING_ACTION_SEND_EMAIL,
            configuration={
                "subject": "Training Completion Reminder",
                "message": "Dear {{ user.first_name }}, please complete the training.",
                "recipients": ["user"],
            },
            user_filter="prospective",  # Prospective users only
        )
        action.full_clean()

        # Test that action applies only to prospective user without NEMO account
        self.assertFalse(action.applies_to_user(self.prospective_staff))
        self.assertFalse(action.applies_to_user(self.prospective_student))
        self.assertTrue(action.applies_to_user(self.prospective_only))

        email_count = EmailLog.objects.count()
        user_training_student = OnlineUserTraining.objects.create(
            online_training=self.training, prospective_user=self.prospective_student
        )
        action_handlers[action.action_type].perform(action, user_training_student)
        self.assertEqual(EmailLog.objects.count(), email_count)
        user_training_staff = OnlineUserTraining.objects.create(
            online_training=self.training, prospective_user=self.prospective_staff
        )
        action_handlers[action.action_type].perform(action, user_training_staff)
        self.assertEqual(EmailLog.objects.count(), email_count)
        user_training_prospective = OnlineUserTraining.objects.create(
            online_training=self.training, prospective_user=self.prospective_only
        )
        action_handlers[action.action_type].perform(action, user_training_prospective)
        self.assertEqual(EmailLog.objects.count(), email_count + 1)

    def test_self_enrollment_training_with_action(self):
        """Test that a user can self-enroll in a training and actions are triggered upon completion."""
        from django.core.signing import TimestampSigner
        from django.urls import reverse

        from NEMO_online_training.models import OnlineUserTraining

        # Create a self-enrollment training with an action
        self_enroll_training = OnlineTraining.objects.create(
            name="Self Enrollment Training",
            completion_time_limit=120,
            enabled=True,
            allow_self_enrollment=True,
        )

        # Create action that applies to all NEMO users
        action = OnlineTrainingAction.objects.create(
            online_training=self_enroll_training,
            action_type=ONLINE_TRAINING_ACTION_EXTEND_ACCESS,
            configuration={"extend_by_days": 90},
            user_filter="all_nemo",
        )

        # Set initial access expiration for student
        initial_date = timezone.now()
        self.student_user.access_expiration = initial_date
        self.student_user.save()

        # Login as student user
        self.login_as(self.student_user)

        # Create training assignment (simulating self-enrollment)
        user_training = OnlineUserTraining.objects.create(
            prospective_user=self.prospective_student,
            online_training=self_enroll_training,
            start=timezone.now(),
        )

        # Verify training was created but not completed
        self.assertIsNone(user_training.end)
        self.assertFalse(user_training.completed())

        # Get the public training URL
        signed_id = TimestampSigner().sign(str(user_training.id))
        training_url = reverse("public_online_training_user_training", args=[signed_id])

        # Access the training page to verify it loads
        response = self.client.get(training_url)
        self.assertEqual(response.status_code, 200)

        # Complete the training via the public completion endpoint
        completion_url = reverse("public_online_training_complete_user_training")
        completion_data = {
            "completion_token": signed_id,
            "answer1": "test answer",
            "score": "95",
        }
        response = self.client.post(completion_url, completion_data)
        self.assertEqual(response.status_code, 200)

        # Verify training was completed
        user_training.refresh_from_db()
        self.assertIsNotNone(user_training.end)
        self.assertTrue(user_training.completed())
        self.assertEqual(user_training.completion_data["answer1"], "test answer")
        self.assertEqual(user_training.completion_data["score"], "95")

        # Verify action was triggered and access was extended
        self.student_user.refresh_from_db()
        expected_date = initial_date + timedelta(days=90)
        self.assertEqual(self.student_user.access_expiration, expected_date.date())

    def test_self_enrollment_training_without_matching_action(self):
        """Test that completing a training doesn't trigger actions that don't apply to the user."""
        from django.core.signing import TimestampSigner
        from django.urls import reverse

        from NEMO_online_training.models import OnlineUserTraining

        # Create a self-enrollment training
        self_enroll_training = OnlineTraining.objects.create(
            name="Self Enrollment Training 2",
            completion_time_limit=120,
            enabled=True,
            allow_self_enrollment=True,
        )

        # Create action that applies only to staff (not student)
        action = OnlineTrainingAction.objects.create(
            online_training=self_enroll_training,
            action_type=ONLINE_TRAINING_ACTION_EXTEND_ACCESS,
            configuration={"extend_by_days": 90},
            user_filter=str(self.staff_type.id),  # Staff only
        )

        # Set initial access expiration for student
        initial_date = timezone.now()
        self.student_user.access_expiration = initial_date
        self.student_user.save()

        # Login as student user
        self.login_as(self.student_user)
        # Go to training page
        response = self.client.get(reverse("online_training_training", args=[self_enroll_training.id]), follow=True)
        self.assertEqual(response.status_code, 200)
        user_training = OnlineUserTraining.objects.get(
            prospective_user=self.prospective_student, online_training=self_enroll_training
        )
        self.assertTrue(user_training)

        # Complete the training via the authenticated endpoint
        signed_id = TimestampSigner().sign(str(user_training.id))
        completion_url = reverse("public_online_training_complete_user_training")
        completion_data = {
            "completion_token": signed_id,
            "answer": "test",
        }
        response = self.client.post(completion_url, completion_data)
        self.assertEqual(response.status_code, 200)

        # Verify training was completed
        user_training.refresh_from_db()
        self.assertTrue(user_training.completed())

        # Verify action was NOT triggered (student's access should remain unchanged)
        self.student_user.refresh_from_db()
        self.assertEqual(self.student_user.access_expiration, initial_date.date())



from NEMO_online_training.apps import OnlineTrainingConfig

ONLINE_TRAINING_NOTIFICATION_TYPE = "online_trainings"
ONLINE_TRAINING_EMAIL_CATEGORY = OnlineTrainingConfig.get_plugin_id() + 1

ONLINE_TRAINING_ACTION_EXTEND_ACCESS = "EXTEND_ACCESS"
ONLINE_TRAINING_ACTION_REMOVE_TRAINING_REQUIRED = "REMOVE_TRAINING_REQUIRED"
ONLINE_TRAINING_ACTION_SEND_EMAIL = "SEND_EMAIL"



import json
from datetime import timedelta
from logging import getLogger
from typing import Optional

from NEMO.decorators import user_office_or_manager_required
from NEMO.models import User
from NEMO.utilities import format_datetime, queryset_search_filter, render_email_template
from NEMO.views.pagination import SortedPaginator
from django.contrib.auth.decorators import login_required
from django.core.signing import BadSignature, SignatureExpired, TimestampSigner
from django.http import HttpResponse, HttpResponseBadRequest, JsonResponse
from django.shortcuts import get_object_or_404, redirect, render
from django.urls import reverse
from django.utils import timezone
from django.utils.functional import Promise
from django.utils.translation import gettext_lazy as _
from django.views.decorators.csrf import ensure_csrf_cookie
from django.views.decorators.http import require_GET, require_POST

from NEMO_online_training.customization import OnlineTrainingCustomization
from NEMO_online_training.forms import OnlineUserTrainingForm, ProspectiveUserForm
from NEMO_online_training.models import OnlineTraining, OnlineUserTraining, ProspectiveUser

online_training_logger = getLogger(__name__)


@login_required
@require_GET
def user_online_trainings(request, prospective_user_id=None):
    user: User = request.user
    selected_status = request.GET.get("training_status", "")
    selected_user_type = request.GET.get("user_type", "")

    current_user_trainings = False
    single_user_view = False

    user_is_staff = user.is_user_office or user.is_facility_manager or user.is_superuser

    prospective_users = ProspectiveUser.objects_with_trainings()
    if not user_is_staff:
        current_user_trainings = True
        single_user_view = True
        prospective_users = prospective_users.filter(nemo_user=user)
    elif prospective_user_id:
        single_user_view = True
        prospective_users = prospective_users.filter(id=prospective_user_id)

    if selected_status == "complete":
        prospective_users = prospective_users.filter(all_trainings_completed=True)
    elif selected_status == "incomplete":
        prospective_users = prospective_users.filter(all_trainings_completed=False)
    if selected_user_type == "new":
        prospective_users = prospective_users.filter(nemo_user__isnull=True)
    elif selected_user_type == "nemo":
        prospective_users = prospective_users.filter(nemo_user__isnull=False)

    page = SortedPaginator(prospective_users, request, order_by="-last_updated").get_current_page()

    # if bool(request.GET.get("csv", False)):
    #     return export_prospective_users(request, prospective_users.order_by("-last_updated"))

    dictionary = {
        "page": page,
        "current_user_trainings": current_user_trainings,
        "user_is_staff": user_is_staff,
        "single_user_view": single_user_view,
        "selected_status": selected_status,
        "selected_user_type": selected_user_type,
        "available_trainings": OnlineTraining.objects.all(),
    }
    return render(request, "NEMO_online_training/user_trainings/user_trainings.html", dictionary)


@require_GET
@user_office_or_manager_required
def search_prospective_users(request):
    return render(request, "NEMO_online_training/user_trainings/user_search.html", {"form": ProspectiveUserForm()})


@user_office_or_manager_required
@require_GET
def prospective_users_search_results(request):
    nemo_users: HttpResponse = queryset_search_filter(
        User.objects.all(), ["first_name", "last_name", "username", "email"], request
    )
    prospective_users: HttpResponse = queryset_search_filter(
        ProspectiveUser.objects.all(), ["_first_name", "_last_name", "_email"], request
    )
    return HttpResponse(
        json.dumps(json.loads(prospective_users.content) + json.loads(nemo_users.content)), "application/json"
    )


@user_office_or_manager_required
@require_GET
def create_prospective_user_from_nemo_user(request, nemo_user_id):
    nemo_user = get_object_or_404(User, pk=nemo_user_id)
    prospective_user = ProspectiveUser.create_from_nemo_user(nemo_user)
    return redirect("online_user_trainings", prospective_user_id=prospective_user.id)


@user_office_or_manager_required
@require_POST
def create_prospective_user(request):
    form = ProspectiveUserForm(request.POST or None)
    if form.is_valid():
        prospective_user = form.save()
        return redirect("online_user_trainings", prospective_user_id=prospective_user.id)
    return render(request, "NEMO_online_training/user_trainings/user_search.html", {"form": form})


@login_required
@require_GET
def create_nemo_user_from_prospective_user(request, prospective_user_id):
    prospective_user = get_object_or_404(ProspectiveUser, pk=prospective_user_id)
    return redirect(
        reverse("create_or_modify_user", kwargs={"user_id": "new"})
        + f"?first_name={prospective_user.first_name}&last_name={prospective_user.last_name}&email={prospective_user.email}&correlation_id={prospective_user_id}"
    )


@login_required
@require_GET
def training_without_assignment(request, online_training_id):
    online_training = get_object_or_404(OnlineTraining, pk=online_training_id)
    if not online_training.allow_self_enrollment:
        return render(
            request,
            "NEMO_online_training/error_message.html",
            {
                "title": "Error",
                "message": "This training is not available for self enrollment, contact staff to be assigned this training",
            },
        )
    if not online_training.enabled:
        return render(
            request,
            "NEMO_online_training/error_message.html",
            {"title": "Error", "message": "This training is not available anymore"},
        )

    prospective_user = ProspectiveUser.create_from_nemo_user(request.user)
    online_user_training, created = OnlineUserTraining.objects.get_or_create(
        prospective_user=prospective_user, online_training=online_training, end=None
    )

    return redirect("online_training_user_training", user_training_id=online_user_training.id)


@require_GET
@login_required
def training(request, user_training_id):
    online_training_user = get_object_or_404(OnlineUserTraining, pk=user_training_id)
    if (
        not online_training_user.prospective_user.nemo_user
        or request.user != online_training_user.prospective_user.nemo_user
    ):
        return render(
            request,
            "NEMO_online_training/error_message.html",
            {"title": "Error", "message": "You do not have permission to complete this training"},
        )

    return redirect(online_training_user.generate_link())


@user_office_or_manager_required
@require_POST
def add_training_to_user(request, prospective_user_id, online_training_id):
    prospective_user = get_object_or_404(ProspectiveUser, pk=prospective_user_id)
    online_training = get_object_or_404(OnlineTraining, pk=online_training_id)

    form = OnlineUserTrainingForm(request.POST)
    form.instance.prospective_user = prospective_user
    form.instance.online_training = online_training

    if form.is_valid():
        online_training_user = form.save()
        online_training_user.generate_and_send_new_email()
        return JsonResponse({"success": True})
    else:
        # Return form errors in a format that JavaScript can handle
        errors = {}
        for field, error_list in form.errors.items():
            errors[field] = error_list
        return JsonResponse({"success": False, "errors": errors}, status=400)


@require_GET
@ensure_csrf_cookie
def public_user_training(request, signed_user_training_id):
    user_training_id = None

    try:
        signer = TimestampSigner()
        max_age = OnlineTrainingCustomization.get_int("online_training_link_validity_minutes") * 60

        # Just check validity
        user_training_id = signer.unsign(signed_user_training_id)
        online_training_user = get_object_or_404(OnlineUserTraining, id=user_training_id)
        if online_training_user.completed():
            return render(
                request,
                "NEMO_online_training/error_message.html",
                {"title": "Success", "message": _("This training has been completed!")},
            )

        # Now check the time limit
        user_training_id = signer.unsign(signed_user_training_id, max_age=max_age)

        # Extract the timestamp manually. Django token format is: value:timestamp:signature
        # We grab the middle part (timestamp) which is Base62 encoded.
        parts = signed_user_training_id.rsplit(":", 2)
        if len(parts) < 3:
            raise BadSignature()
    except (BadSignature, SignatureExpired) as e:
        if isinstance(e, SignatureExpired) and request.user and request.user.is_authenticated:
            return redirect("online_training_user", user_training_id=user_training_id)
        return render(
            request,
            "NEMO_online_training/public/invalid_training_link.html",
            {"user_training_id": user_training_id},
        )

    online_training_user.prospective_user.last_accessed = timezone.now()
    online_training_user.prospective_user.save(update_fields=["last_accessed"])
    error = check_training_validity(online_training_user)
    if error:
        return render(request, "NEMO_online_training/error_message.html", {"title": "Error", "message": error})
    else:
        completion_token = TimestampSigner().sign(user_training_id)
        online_training_user.start = timezone.now()
        online_training_user.save(update_fields=["start"])

        training_context = {
            "training_user": online_training_user.prospective_user,
            "training": online_training_user.online_training,
            "record": online_training_user,
        }

        online_training_rendered = render_email_template(
            online_training_user.online_training.html_content, training_context, request
        )
        return render(
            request,
            "NEMO_online_training/public/user_training.html",
            {
                "online_training_user": online_training_user,
                "online_training_rendered": online_training_rendered,
                "expires_at": online_training_user.start
                + timedelta(minutes=online_training_user.online_training.completion_time_limit),
                "completion_token": completion_token,
            },
        )


@require_POST
def public_generate_user_training_email(request, user_training_id):
    online_training_user = OnlineUserTraining.objects.filter(id=user_training_id).first()
    if not online_training_user:
        if request.POST.get("popup"):
            return HttpResponseBadRequest(_("Invalid link"))
        else:
            return render(
                request,
                "NEMO_online_training/error_message.html",
                {"title": "Error", "message": _("Invalid link")},
                status=400,
            )
    online_training_user.generate_and_send_new_email()
    return render(request, "NEMO_online_training/public/new_link_email_confirmation.html")


@require_POST
def public_complete_user_training(request):
    signed_user_training_id = request.POST.get("completion_token")
    user_training_id = None

    try:
        signer = TimestampSigner()
        # Just check validity
        user_training_id = signer.unsign(signed_user_training_id)
        # Now check the time limit
        online_training_user = get_object_or_404(OnlineUserTraining, pk=user_training_id)
        dynamic_limit_seconds = online_training_user.online_training.completion_time_limit * 60
        signer.unsign(signed_user_training_id, max_age=dynamic_limit_seconds)
    except (BadSignature, SignatureExpired):
        return render(
            request, "NEMO_online_training/public/invalid_training_link.html", {"user_training_id": user_training_id}
        )

    error = check_training_validity(online_training_user)
    if error:
        return render(request, "NEMO_online_training/error_message.html", {"title": "Error", "message": error})
    else:
        data = {}
        for key, values in request.POST.lists():
            # remove completion token and csrf token from the completion data
            if key in ["csrfmiddlewaretoken", "completion_token"]:
                continue
            if len(values) == 1:
                data[key] = values[0]
            else:
                data[key] = values
        online_training_user.complete(data)

    return HttpResponse()


def check_training_validity(online_user_training: OnlineUserTraining) -> Optional[Promise | str]:
    if not online_user_training.online_training.enabled:
        return _("This training is not available anymore!")
    if online_user_training.has_training_expired():
        return _(f"This training expired on {format_datetime(online_user_training.due_date)}")
    if online_user_training.end:
        return _("This training has been completed!")
    return None



from typing import List, Optional, Tuple, Union

from NEMO.exceptions import NEMOException, ProjectChargeException, UserAccessError
from NEMO.models import (
    Area,
    AreaAccessRecord,
    Consumable,
    ConsumableWithdraw,
    Project,
    Reservation,
    StaffCharge,
    Tool,
    UsageEvent,
    User,
)
from NEMO.policy import BaseNEMOPolicy
from django.db.models import QuerySet
from django.http import HttpResponse, HttpResponseBadRequest
from django.utils.safestring import mark_safe

from NEMO_online_training.models import OnlineUserTraining, ProspectiveUser


class BlockingTrainingDueException(NEMOException):
    pass


class ProjectBlockingTrainingDueException(ProjectChargeException):
    pass


class AccessBlockingTrainingDueException(UserAccessError):
    pass


class NEMOOnlineTrainingPolicy(BaseNEMOPolicy):

    def check_to_save_reservation(
        self,
        cancelled_reservation: Optional[Reservation],
        new_reservation: Reservation,
        user_creating_reservation: User,
        explicit_policy_override: bool,
    ) -> Tuple[List[str], bool]:
        policy_problems = []

        if new_reservation.user.is_staff_on_tool(new_reservation.tool) or explicit_policy_override:
            return policy_problems, False

        prospective_user = ProspectiveUser.objects.filter(nemo_user=new_reservation.user).first()
        if prospective_user and not prospective_user.all_blocking_trainings_completed():
            msg = get_blocking_training_error_message(prospective_user.all_blocking_trainings_due())
            policy_problems.append(msg)
            return policy_problems, True

        return policy_problems, True

    def check_to_enable_tool(
        self, tool: Tool, operator: User, user: User, project: Project, staff_charge: bool, remote_work=False
    ) -> HttpResponse:
        if not operator.is_staff_on_tool(tool):
            prospective_user = ProspectiveUser.objects.filter(nemo_user=user).first()
            if prospective_user and not prospective_user.all_blocking_trainings_completed():
                msg = get_blocking_training_error_message(prospective_user.all_blocking_trainings_due())
                return HttpResponseBadRequest(msg)
        return HttpResponse()

    def check_to_enter_any_area(self, user: User):
        prospective_user = ProspectiveUser.objects.filter(nemo_user=user).first()
        if prospective_user and not prospective_user.all_blocking_trainings_completed():
            msg = get_blocking_training_error_message(prospective_user.all_blocking_trainings_due(), html=False)
            raise AccessBlockingTrainingDueException(msg)

    def check_billing_to_project(
        self,
        project: Project,
        user: User,
        item: Union[Tool, Area, Consumable, StaffCharge] = None,
        charge: Union[UsageEvent, AreaAccessRecord, ConsumableWithdraw, StaffCharge, Reservation] = None,
    ):
        if not isinstance(charge, (Reservation, UsageEvent, AreaAccessRecord)):
            prospective_user = ProspectiveUser.objects.filter(nemo_user=user).first()
            if prospective_user and not prospective_user.all_blocking_trainings_completed():
                msg = get_blocking_training_error_message(prospective_user.all_blocking_trainings_due())
                raise ProjectBlockingTrainingDueException(project=project, user=user, msg=msg)


def get_blocking_training_error_message(user_trainings_due: QuerySet[OnlineUserTraining], html=True) -> str:
    training_names = user_trainings_due.values_list("online_training__name", flat=True)
    training_list = ""
    if html:
        for training in training_names:
            training_list += f"<li style='margin-left: 40px'>{training}</li>"
        return mark_safe(
            f"<ul style='padding-left:0'>The following trainings must be completed before proceeding: {training_list}</ul>"
        )
    else:
        return f"The following trainings must be completed before proceeding: {', '.join(training_names)}"



from abc import ABC, abstractmethod
from typing import Dict

from NEMO.utilities import render_email_template
from NEMO.views.customization import ApplicationCustomization
from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _

from NEMO_online_training.fields import UserTypeFilterField
from NEMO_online_training.utilities import (
    ONLINE_TRAINING_ACTION_EXTEND_ACCESS,
    ONLINE_TRAINING_ACTION_REMOVE_TRAINING_REQUIRED,
    ONLINE_TRAINING_ACTION_SEND_EMAIL,
)


class OnlineTrainingActionHandler(ABC):
    """
    Base class for all online training action handlers.
    Similar to NEMO's interlock pattern.
    """

    @abstractmethod
    def validate(self, configuration: dict, user_filter: list[str]):
        """
        Validate the action configuration and user filter.
        Raise ValidationError if the configuration or user filter is invalid.

        Args:
            configuration: The JSON configuration dict from OnlineTrainingAction
            user_filter: The list of user types to apply the action to
        """
        if not isinstance(configuration, dict):
            raise ValidationError(_("Configuration must be a dictionary"))

    def perform(self, action, user_training) -> None:
        if action.applies_to_user(user_training.prospective_user):
            self.do_perform(action, user_training)

    @abstractmethod
    def do_perform(self, action, user_training) -> None:
        """
        Perform the action.

        Args:
            action: The OnlineTrainingAction
            user_training: The OnlineUserTraining
        """
        pass

    @property
    @abstractmethod
    def name(self) -> str:
        """
        Return the name of the action to be saved in the database.
        """
        pass

    @property
    @abstractmethod
    def description(self) -> str:
        """
        Return a description of the action to be shown to the user.
        """
        pass


class ExtendAccessOnlineTrainingHandler(OnlineTrainingActionHandler):
    """Handler for extending user access expiration"""

    @property
    def name(self) -> str:
        return ONLINE_TRAINING_ACTION_EXTEND_ACCESS

    @property
    def description(self) -> str:
        return _("Extend User Access Expiration")

    def validate(self, configuration: dict, user_filter: list[str]) -> None:
        super().validate(configuration, user_filter)

        if UserTypeFilterField.PROSPECTIVE_USERS in user_filter:
            raise ValidationError({"user_filter": _("New users cannot have their access extended")})

        if "extend_by_days" not in configuration:
            raise ValidationError({"configuration": _("Configuration must include 'extend_by_days' field")})

        extend_by_days = configuration.get("extend_by_days")
        if not isinstance(extend_by_days, (int, float)) or extend_by_days <= 0:
            raise ValidationError({"configuration": _("'extend_by_days' must be a positive number")})

    def do_perform(self, action, user_training) -> None:
        from datetime import timedelta
        from django.utils import timezone

        extend_by_days = action.configuration["extend_by_days"]

        # If the user is linked to a NEMO user
        if user_training.prospective_user.nemo_user:
            nemo_user = user_training.prospective_user.nemo_user
            current_expiration = nemo_user.access_expiration or timezone.now()
            nemo_user.access_expiration = current_expiration + timedelta(days=extend_by_days)
            nemo_user.save(update_fields=["access_expiration"])


class RemoveTrainingRequiredOnlineTrainingHandler(OnlineTrainingActionHandler):
    """Handler for removing training requirement on user"""

    @property
    def name(self) -> str:
        return ONLINE_TRAINING_ACTION_REMOVE_TRAINING_REQUIRED

    @property
    def description(self) -> str:
        return _("Remove training required on user account")

    def validate(self, configuration: dict, user_filter: list[str]) -> None:
        super().validate(configuration, user_filter)

        if UserTypeFilterField.PROSPECTIVE_USERS in user_filter:
            raise ValidationError(
                {
                    "user_filter": _(
                        f"New users cannot have their {ApplicationCustomization.get('facility_rules_name')} requirement removed"
                    )
                }
            )

    def do_perform(self, action, user_training) -> None:
        # If the user is linked to a NEMO user and has training required
        if user_training.prospective_user.nemo_user and user_training.online_training.training_required:
            nemo_user = user_training.prospective_user.nemo_user
            nemo_user.training_required = False
            nemo_user.save(update_fields=["training_required"])


class SendEmailOnlineTrainingHandler(OnlineTrainingActionHandler):
    """Handler for sending notification emails"""

    @property
    def name(self) -> str:
        return ONLINE_TRAINING_ACTION_SEND_EMAIL

    @property
    def description(self) -> str:
        return _("Send Notification Email")

    def validate(self, configuration: dict, user_filter: list) -> None:
        super().validate(configuration, user_filter)

        if "subject" not in configuration:
            raise ValidationError({"configuration": _("Configuration must include 'subject' field")})

        if "message" not in configuration:
            raise ValidationError({"configuration": _("Configuration must include 'message' field")})

        if "recipients" not in configuration:
            raise ValidationError({"configuration": _("Configuration must include 'recipients' field")})

        recipients = configuration.get("recipients")
        if not isinstance(recipients, list) or not recipients:
            raise ValidationError({"configuration": _("'recipients' must be a non-empty list")})

        valid_recipient_types = ["user"]
        for recipient in recipients:
            if not isinstance(recipient, str):
                raise ValidationError({"configuration": _("Each recipient must be a string")})
            # Check if it's a valid type or an email
            if recipient not in valid_recipient_types and "@" not in recipient:
                raise ValidationError(
                    {"configuration": _(f"Invalid recipient '{recipient}'. Must be 'user', or a valid email address")}
                )

    def do_perform(self, action, user_training) -> None:
        from NEMO.utilities import send_mail

        from NEMO_online_training.utilities import ONLINE_TRAINING_EMAIL_CATEGORY

        subject = action.configuration["subject"]
        message = action.configuration["message"]
        recipients = action.configuration["recipients"]

        # Format message with available context
        context = {
            "training_user": user_training.prospective_user,
            "training": user_training.online_training,
            "record": user_training,
            "action": action,
        }
        formatted_message = render_email_template(message, context)
        formatted_subject = render_email_template(subject, context)

        # Build recipient list
        recipient_emails = []
        for recipient in recipients:
            if recipient == "user":
                recipient_emails.append(user_training.prospective_user.email)
            else:
                # Assume it's an email address
                recipient_emails.append(recipient)

        if recipient_emails:
            send_mail(
                formatted_subject,
                formatted_message,
                from_email=None,
                to=recipient_emails,
                email_category=ONLINE_TRAINING_EMAIL_CATEGORY,
            )


# Registry of all action handlers
action_handlers: Dict[str, OnlineTrainingActionHandler] = {
    ONLINE_TRAINING_ACTION_EXTEND_ACCESS: ExtendAccessOnlineTrainingHandler(),
    ONLINE_TRAINING_ACTION_REMOVE_TRAINING_REQUIRED: RemoveTrainingRequiredOnlineTrainingHandler(),
    ONLINE_TRAINING_ACTION_SEND_EMAIL: SendEmailOnlineTrainingHandler(),
}



from NEMO.decorators import customization
from NEMO.views.customization import CustomizationBase


@customization(key="online_training", title="Online training")
class OnlineTrainingCustomization(CustomizationBase):
    variables = {
        "online_training_feature_name": "Online trainings",
        "online_training_link_validity_minutes": "7200",  # 5 days
        "online_training_user_unique_email": "enabled",
    }

    files = [
        ("online_training_send_new_link_email", ".html"),
    ]

    def __init__(self, key, title):
        super().__init__(key, title)
        self.title = self.get("online_training_feature_name", raise_exception=False, use_cache=False)



from datetime import timedelta

from NEMO.constants import CHAR_FIELD_MEDIUM_LENGTH, CHAR_FIELD_SMALL_LENGTH
from NEMO.models import BaseModel, Notification, SerializationByNameModel, User
from NEMO.utilities import format_datetime, format_timedelta, get_full_url, render_email_template, send_mail
from NEMO.views.customization import get_media_file_contents
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import NON_FIELD_ERRORS, ValidationError
from django.core.signing import TimestampSigner
from django.db import models, transaction
from django.db.models import BooleanField, Case, Count, F, Q, When
from django.db.models.signals import post_save, pre_delete
from django.dispatch import receiver
from django.urls import reverse
from django.utils import timezone
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _

from NEMO_online_training.customization import OnlineTrainingCustomization
from NEMO_online_training.fields import UserTypeFilterField
from NEMO_online_training.utilities import ONLINE_TRAINING_EMAIL_CATEGORY, ONLINE_TRAINING_NOTIFICATION_TYPE


class ProspectiveUser(BaseModel):
    creation_time = models.DateTimeField(auto_now_add=True)
    last_updated = models.DateTimeField(auto_now=True)
    last_accessed = models.DateTimeField(null=True, blank=True)
    _first_name = models.CharField(
        verbose_name="First name", db_column="first_name", null=True, blank=True, max_length=CHAR_FIELD_SMALL_LENGTH
    )
    _last_name = models.CharField(
        verbose_name="Last name", db_column="last_name", null=True, blank=True, max_length=CHAR_FIELD_SMALL_LENGTH
    )
    _email = models.EmailField(verbose_name="Email address", db_column="email", null=True, blank=True)
    nemo_user = models.OneToOneField(User, on_delete=models.SET_NULL, null=True, blank=True)

    class Meta:
        ordering = ["_first_name", "_last_name", "nemo_user__first_name", "nemo_user__last_name"]

    @property
    def first_name(self):
        return self._first_name or self.nemo_user.first_name

    @first_name.setter
    def first_name(self, value):
        self._first_name = value

    @property
    def last_name(self):
        return self._last_name or self.nemo_user.last_name

    @last_name.setter
    def last_name(self, value):
        self._last_name = value

    @property
    def email(self):
        return self._email or self.nemo_user.email

    @email.setter
    def email(self, value):
        self._email = value

    def all_trainings_completed(self):
        return not self.onlineusertraining_set.filter(end__isnull=True).exists()

    def all_blocking_trainings_completed(self):
        """
        Blocking trainings are those that are marked as blocking and have a due date in the past.
        """
        now = timezone.now()
        return not self.onlineusertraining_set.filter(
            online_training__is_blocking=True, due_date__lt=now, end__isnull=True
        ).exists()

    def all_blocking_trainings_due(self):
        return self.onlineusertraining_set.filter(
            online_training__is_blocking=True, due_date__lt=timezone.now(), end__isnull=True
        )

    def get_name(self):
        return self.first_name + " " + self.last_name

    def all_incomplete_training_ids(self):
        return list(self.onlineusertraining_set.filter(end__isnull=True).values_list("id", flat=True))

    @staticmethod
    def objects_with_trainings():
        return (
            ProspectiveUser.objects.annotate(
                # Count all trainings linked to this user
                total_trainings=Count("onlineusertraining"),
                # Count only the trainings that have a completed date
                completed_trainings=Count("onlineusertraining", filter=Q(onlineusertraining__end__isnull=False)),
            )
            .annotate(
                # Compare the two counts to create the boolean
                # Logic: total == completed AND total > 0
                all_trainings_completed=Case(
                    When(Q(total_trainings=F("completed_trainings")) & Q(total_trainings__gt=0), then=True),
                    default=False,
                    output_field=BooleanField(),
                )
            )
            .prefetch_related("onlineusertraining_set")
        )

    @staticmethod
    def create_from_nemo_user(nemo_user: User):
        prospective_user, created = ProspectiveUser.objects.get_or_create(nemo_user=nemo_user)
        if created:
            prospective_user.save()
        return prospective_user

    def clean(self):
        if OnlineTrainingCustomization.get_bool("online_training_user_unique_email"):
            if ProspectiveUser.objects.filter(_email=self.email).exclude(id=self.id).exists():
                raise ValidationError({"email": _("This email is already used by another user.")})

    def save(self, *args, **kwargs):
        if self.nemo_user:
            self._first_name = None
            self._last_name = None
            self._email = None
        super().save(*args, **kwargs)

    def __str__(self):
        return f"{self.first_name} {self.last_name}"


class OnlineTraining(SerializationByNameModel):
    name = models.CharField(max_length=CHAR_FIELD_MEDIUM_LENGTH, unique=True)
    enabled = models.BooleanField(
        default=True, help_text=_("If unchecked, this training will not be available to users")
    )
    completion_time_limit = models.PositiveIntegerField(
        default=120, help_text=_("The maximum time the user can take to complete the training on the page (in minutes)")
    )
    is_blocking = models.BooleanField(
        default=False,
        help_text=_(
            "If checked, prevents the linked NEMO user from performing other actions until this training is completed"
        ),
    )
    allow_self_enrollment = models.BooleanField(
        default=False, help_text=_("Allow users to take this training without prior assignment.")
    )
    html_content = models.TextField(
        null=True,
        blank=True,
        help_text=mark_safe(
            _(
                "The HTML content of the training. The following context variables are available:"
                "<ul style='padding-left: 35px;'>"
                "<li style='list-style: initial'><b>training_user</b>: the user who is completing the training</li>"
                "<li style='list-style: initial'><b>training</b>: the training being completed</li>"
                "<li style='list-style: initial'><b>record</b>: the completion record</li>"
                "</ul>"
                "Upon completion, call the JS function: <code>training_completed(dict_data)</code>"
            )
        ),
    )
    creation_time = models.DateTimeField(auto_now_add=True)

    class Meta:
        ordering = ["name"]

    def __str__(self):
        return self.name


class OnlineTrainingAction(BaseModel):
    online_training = models.ForeignKey(OnlineTraining, on_delete=models.CASCADE)
    action_type = models.CharField(max_length=CHAR_FIELD_SMALL_LENGTH)
    configuration = models.JSONField(
        default=dict,
        blank=True,
        help_text=_(
            "Enter configuration as JSON. See <a href='https://github.com/usnistgov/NEMO-online-training?tab=readme-ov-file#usage' target='_blank'>this link</a> for more information."
        ),
    )
    user_filter = UserTypeFilterField(
        help_text=_(
            "Select which users this action applies to. You can select 'All NEMO users', "
            "specific user types, and/or new users (without NEMO accounts)."
        ),
        default="all_nemo,prospective",
    )

    class Meta:
        ordering = ["action_type"]

    def clean(self):
        from NEMO_online_training.training_actions import action_handlers

        if self.action_type not in action_handlers:
            raise ValidationError(_(f"Invalid action type: {self.action_type}"))

        handler = action_handlers[self.action_type]
        handler.validate(self.configuration, self.user_filter)

    def applies_to_user(self, prospective_user) -> bool:
        field = UserTypeFilterField()
        return field.applies_to_user(self.user_filter, prospective_user)

    def __str__(self):
        return f"{self.action_type} for {self.online_training.name}"


class OnlineUserTraining(BaseModel):
    online_training = models.ForeignKey(OnlineTraining, on_delete=models.CASCADE)
    prospective_user = models.ForeignKey(ProspectiveUser, on_delete=models.CASCADE)
    due_date = models.DateTimeField(null=True, blank=True, help_text=_("The due date/time for the training"))
    start = models.DateTimeField(null=True, blank=True, help_text=_("The date/time the training was started"))
    end = models.DateTimeField(null=True, blank=True, help_text=_("The date/time the training was completed"))
    completion_data = models.JSONField(
        null=True, blank=True, help_text=_("Completion data to be stored for the training, like answers to questions")
    )
    creation_time = models.DateTimeField(auto_now_add=True)
    last_updated = models.DateTimeField(auto_now=True)

    class Meta:
        ordering = ["-end", "-due_date"]

    def has_training_expired(self) -> bool:
        return self.due_date and self.due_date < timezone.now()

    def generate_link(self) -> str:
        return get_full_url(
            reverse("public_online_training_user_training", args=[TimestampSigner().sign(str(self.id))])
        )

    def generate_and_send_new_email(self):
        new_link_template = get_media_file_contents("online_training_send_new_link_email.html")
        if not new_link_template:
            new_link_template = "Dear {{ training_user.first_name }},<br><br>Please click on the following unique link to complete your assigned training: <br><a href='{{ record.generate_link }}'>complete training</a>."
        message = render_email_template(
            new_link_template,
            {"training_user": self.prospective_user, "training": self.online_training, "record": self},
        )
        send_mail(
            subject=f"Link to complete {self.online_training.name}",
            content=message,
            from_email=None,
            to=[self.prospective_user.email],
            email_category=ONLINE_TRAINING_EMAIL_CATEGORY,
        )

    def completed(self):
        return self.end is not None

    def completion_time(self) -> timedelta | str | None:
        if self.start and self.end:
            return format_timedelta(self.end - self.start, "{H:02}h {M:02}m {S:02}s")
        elif self.start:
            return "ongoing"
        return None

    def complete(self, data: dict = None):
        from NEMO_online_training.training_actions import action_handlers

        self.end = timezone.now()
        self.completion_data = data
        self.save()
        for action in self.online_training.onlinetrainingaction_set.all():
            handler = action_handlers[action.action_type]
            handler.perform(action, self)

    def clean(self):
        if self.prospective_user and self.online_training:
            # Check for duplicate incomplete trainings
            if (
                OnlineUserTraining.objects.filter(
                    prospective_user=self.prospective_user,
                    online_training=self.online_training,
                    end__isnull=True,
                    due_date__gte=timezone.now(),
                )
                .exclude(id=self.id)
                .exists()
            ):
                raise ValidationError(
                    {
                        NON_FIELD_ERRORS: _(
                            "This user already has an incomplete training of this type. Please complete or remove it first."
                        )
                    }
                )

    def __str__(self):
        due_date = f", due {format_datetime(self.due_date, 'SHORT_DATETIME_FORMAT')}" if self.due_date else ""
        return f"{self.online_training.name} - {self.prospective_user.get_name()}{due_date}"


def notification_qs_for_training(training: OnlineUserTraining):
    ct = ContentType.objects.get_for_model(OnlineUserTraining)
    return Notification.objects.filter(
        notification_type=ONLINE_TRAINING_NOTIFICATION_TYPE, content_type=ct, object_id=training.pk
    )


@receiver(post_save, sender=OnlineUserTraining)
def online_training_user_training_notification_on_save(sender, instance: OnlineUserTraining, created: bool, **kwargs):
    # always remove any previous notifications
    def _delete():
        notification_qs_for_training(instance).delete()

    transaction.on_commit(_delete)

    if instance.prospective_user.nemo_user and not instance.end:
        # Create/Recreate a notification when the training is created or updated (except when it's completed).
        def _update():
            Notification.objects.get_or_create(
                user=instance.prospective_user.nemo_user,
                notification_type=ONLINE_TRAINING_NOTIFICATION_TYPE,
                content_type=ContentType.objects.get_for_model(OnlineUserTraining),
                object_id=instance.id,
                defaults={"expiration": (instance.due_date or timezone.now()) + timedelta(days=30)},
            )

        transaction.on_commit(_update)
        return


@receiver(pre_delete, sender=OnlineUserTraining)
def online_training_user_training_notification_on_delete(sender, instance: OnlineUserTraining, **kwargs):
    def _delete():
        notification_qs_for_training(instance).delete()

    transaction.on_commit(_delete)


@receiver(post_save, sender=User)
def finalize_prospective_user_to_nemo_user_conversion(sender, instance, created, **kwargs):
    # Look for the temporary attribute we set in the view
    corr_id = getattr(instance, "_correlation_id", None)

    if created and corr_id:
        # Link the ProspectiveUser to this new actual User
        prospective_user = ProspectiveUser.objects.filter(id=corr_id).first()
        if prospective_user:
            prospective_user.nemo_user = instance
            prospective_user.save()
